import unittest
import petri
import sample
import helper

Marking = petri.Marking


class TransitionUnitTest(unittest.TestCase):
  """Test the GetUniformTransition()"""

  def testGetUniformTransition(self):
    deltas = [(0, -1), (1, 0)]
    transition = petri.GetUniformTransition(deltas)

    # Test all positive.
    expected = [(1, 0), (2, 1)]
    tuples = [m.tuple for m in transition(Marking((1,1)))]
    self.assertEqual(tuples, expected)

    # Test not all positive.
    expected = [(1, 0)]
    tuples = [m.tuple for m in transition(Marking((0,0)))]
    self.assertEqual(tuples, expected)

class NodeUnitTest(unittest.TestCase):
  """Test the Node class."""

  def testMinDominated(self):
    node = petri.Node(Marking((1,1)), None)

    small = Marking((0,0))
    middle = Marking((2,0))
    large = Marking((2,2))

    mark = Marking((1,0))
    node.min_ancestors = [small, middle]
    self.assertTrue(node.IsMinDominated(mark))

    mark = Marking((1,0))
    node.min_ancestors = [large, middle]
    self.assertFalse(node.IsMinDominated(mark))

    mark = Marking((1,2))
    node.min_ancestors = [large, middle]
    self.assertTrue(node.IsMinDominated(mark))

  def testAddDescendant(self):
    node = petri.Node(None, None)

    node.AddMinAncestor(Marking((1,1)))
    tuples = set(m.tuple for m in node.min_ancestors)
    self.assertEqual(tuples, set([(1,1)]))

    node.AddMinAncestor(Marking((2,1)))
    tuples = set(m.tuple for m in node.min_ancestors)
    self.assertEqual(tuples, set([(1,1)]))

    node.AddMinAncestor(Marking((0,1)))
    tuples = set(m.tuple for m in node.min_ancestors)
    self.assertEqual(tuples, set([(0,1)]))

    node.AddMinAncestor(Marking((1,0)))
    tuples = set(m.tuple for m in node.min_ancestors)
    self.assertEqual(tuples, set([(0,1),(1,0)]))

  def testIsMinDominated2(self):
    node = petri.Node(Marking((2,0)), None)
    node.AddMinAncestor(Marking((0,1)))

    self.assertTrue(node.IsMinDominated(Marking((2,1))))

    #self.assertFalse(node.IsDominated((2,0)))

    self.assertTrue(node.IsMinDominated(Marking((1,1))))

    #self.assertTrue(node.IsDominated((0,1)))

    self.assertFalse(node.IsMinDominated(Marking((0,0))))

  def testPathToRoot(self):
    root = petri.Node(Marking((0,)), None)

    node = petri.Node(Marking((1,)), root)
    root.AddChild(node)

    node2 = petri.Node(Marking((2,)), node)
    node.AddChild(node2)

    self.assertEqual(list(root.pathFromRoot), [root])
    self.assertEqual(list(node.pathFromRoot), [root, node])
    self.assertEqual(list(node2.pathFromRoot), [root, node, node2])

class CoverSetUnitTest(unittest.TestCase):
  """Test the CoverSet class."""

  def testInit(self):
    transition = lambda t: [t.AddTuple((1,0)), t.AddTuple((0,-1))]
    cover = petri.CoverSet(transition, Marking((1,1)))

    values = [c.value.tuple for c in cover.candidates]
    self.assertEqual(values, [(2,1), (1,0)])

  def testCheckDomination(self):
    addCan_orig = petri.CoverSet._AddCandidates
    petri.CoverSet._AddCandidates = lambda *_: None

    try:
      cover = petri.CoverSet(None, Marking((1,1)))
      node = petri.Node(Marking((0,0)), cover.root)
      self.assertTrue(cover._CheckDomination(node))

      node = petri.Node(Marking((1,1)), cover.root)
      self.assertTrue(cover._CheckDomination(node))

      node = petri.Node(Marking((0,2)), cover.root)
      self.assertFalse(cover._CheckDomination(node))

    finally:
      petri.CoverSet._AddCandidates = addCan_orig

  def testProcessAncestor(self):
    orig_path = petri.Node._PathToRoot
    acc = []
    try:


      # 1 dominated ancestor.
      root = petri.Node(Marking((1,0)), None)
      node = petri.Node(Marking((1,1)), None)
      acc, not_anc = petri.CoverSet._ProcessAncestors(root, node)
      self.assertEqual(acc, root)
      self.assertEqual(not_anc, [acc])

      # 1 incomparable ancestor.
      root = petri.Node(Marking((2,0)), None)
      sibling = petri.Node(Marking((3,2)), root)
      node = petri.Node(Marking((1,1)), root)
      root.AddChild(node)
      root.AddChild(sibling)
      acc, not_anc = petri.CoverSet._ProcessAncestors(root, node)
      self.assertEqual(acc, None)
      self.assertEqual(not_anc, [sibling])

      # 2 dominated ancestors
      root = petri.Node(Marking((0,1)), None)
      node2 = petri.Node(Marking((1,0)), root)
      node = petri.Node(Marking((1,1)), node2)
      root.AddChild(node2)
      node2.AddChild(node)
      acc, not_anc = petri.CoverSet._ProcessAncestors(root, node)
      self.assertEqual(acc, root)
      self.assertEqual(not_anc, [acc])

      # incomparable then dominated.
      root = petri.Node(Marking((2,0)), None)
      node2 = petri.Node(Marking((0,1)), root)
      sibling = petri.Node(Marking((3,2)), root)
      node = petri.Node(Marking((1,1)), node2)
      root.AddChild(node2)
      root.AddChild(sibling)
      node2.AddChild(node)
      acc, not_anc = petri.CoverSet._ProcessAncestors(root, node)
      self.assertEqual(acc, node2)
      self.assertEqual(not_anc, [sibling, acc])

    finally:
      petri.Node._PathToRoot = orig_path

  def testProcessNext(self):
    """Test a realist scenario."""
    transition = petri.GetUniformTransition([(-1,1,0,0,0,0,0),
                                             (-1,0,0,0,0,1,0),
                                             (-1,0,0,0,0,0,1),
                                             (0,-1,1,0,0,0,0),
                                             (0,0,-1,1,0,0,0),
                                             (0,1,0,0,1,0,-1),
                                             (0,0,1,-1,1,0,0),
                                             (0,0,0,1,2,-1,0)])
    initial = Marking((1,0,0,0,0,0,0))
    cover = petri.CoverSet(transition, initial)

    vals = [c.value.tuple for c in cover.candidates]
    self.assertEqual(vals, [(0,1,0,0,0,0,0),
                            (0,0,0,0,0,1,0),
                            (0,0,0,0,0,0,1)])

    node = cover.candidates.pop()
    cover.ProcessNext(node)

    vals = [c.value.tuple for c in cover.candidates]
    self.assertEqual(vals, [(0,1,0,0,0,0,0),
                            (0,0,0,0,0,1,0),
                            (0,1,0,0,1,0,0)])
    self.assertEqual(cover.root.children, set([node]))

    node2 = cover.candidates.pop(0)
    cover.ProcessNext(node2)

    vals = [c.value.tuple for c in cover.candidates]
    self.assertEqual(vals, [(0,0,0,0,0,1,0),
                            (0,1,0,0,1,0,0),
                            (0,0,1,0,0,0,0)])
    self.assertEqual(cover.root.children, set([node, node2]))

    node3 = cover.candidates.pop()
    cover.ProcessNext(node3)
    node4 = cover.candidates.pop(1)
    cover.ProcessNext(node4)

    vals = [c.value.tuple for c in cover.candidates]
    self.assertEqual(vals, [(0,0,0,0,0,1,0),
                            (0,0,0,1,0,0,0),
                            (0,0,1,0,1,0,0)])
    self.assertEqual(cover.root.children, set([node]))
    self.assertEqual(node.children, set([node4]))
    self.assertTrue(cover.candidates[1].parent.dead)

    node35 = cover.candidates.pop()
    cover.ProcessNext(node35)
    node45 = cover.candidates.pop()
    cover.ProcessNext(node45)
    node35ax= cover.candidates.pop()
    cover.ProcessNext(node35ax)

  def testCompute(self):
    transition = petri.GetUniformTransition([(-1,1,0,0,0,0,0),
                                             (-1,0,0,0,0,1,0),
                                             (-1,0,0,0,0,0,1),
                                             (0,-1,1,0,0,0,0),
                                             (0,0,-1,1,0,0,0),
                                             (0,1,0,0,1,0,-1),
                                             (0,0,1,-1,1,0,0),
                                             (0,0,0,1,2,-1,0)])
    initial = Marking((1,0,0,0,0,0,0))
    cover = petri.CoverSet(transition, initial)
    cover.Compute()

#  def testProcessNonAncestor(self):
#    root = petri.Node(Marking((0,3)), None)
#    node1 = petri.Node(Marking((2,0)), root)
#    root.AddChild(node1)
#    node2 = petri.Node(Marking((1,2)), root)
#    root.AddChild(node2)
#    node11 = petri.Node(Marking((1,1)), node1)
#    node1.AddChild(node11)
#    node21 = petri.Node(Marking((0,0)), node2)
#    node2.AddChild(node21)
#
#    node = petri.Node(Marking((3,0)), None)
#
#    petri.CoverSet._ProcessNonAncestors(node, [root])
#
#    self.assertFalse(root.dead)
#    self.assertTrue(node1.dead)
#    self.assertTrue(node11.dead)
#    self.assertFalse(node2.dead)
#    self.assertTrue(node21.dead)


class UtilUnitTest(unittest.TestCase):

  def testPairwise(self):
    it = xrange(3)
    pairs = list(petri.Util.Pairwise(it))
    self.assertEqual(pairs, [(0,1),(1,2)])

class MarkingUnitTest(unittest.TestCase):

  def testAccelerate(self):
    mark = Marking((1,1))
    high_Mark = Marking((2,1))
    acc = mark.Accelerate(high_Mark)
    self.assertEqual(acc.tuple, (petri.INF, 1))

    high_Mark = Marking((2,2))
    acc = mark.Accelerate(high_Mark)
    self.assertEqual(acc.tuple, (petri.INF, petri.INF))

  def testOmega(self):
    mark1 = petri.Marking((0,petri.INF))
    mark2 = petri.Marking((1,0))
    mark3 = petri.Marking((petri.INF,0))
    mark4 = petri.Marking((1,petri.INF))

    # Equality.
    self.assertEqual(mark1, mark1)
    self.assertTrue(mark1 == mark1)
    self.assertTrue(mark1 != mark2)
    self.assertFalse(mark1 == mark2)

    # Inequalities.
    self.assertTrue(mark1 <= mark1)
    self.assertTrue(mark2 < mark3)
    self.assertFalse(mark1 < mark2)
    self.assertFalse(mark2 < mark1)
    self.assertFalse(mark3 < mark1)
    self.assertFalse(mark1 < mark3)
    self.assertTrue(mark1 < mark4)

  def testAdd(self):
    mark1 = petri.Marking((0,petri.INF))
    self.assertEqual(mark1.AddTuple((1,1)).tuple, (1, petri.INF))


  def testEquality(self):
    mark1 = petri.Marking((0,0))
    mark2 = petri.Marking((0,0))
    self.assertEqual(mark1, mark1)
    self.assertEqual(mark1, mark2)
    self.assertTrue(mark1 == mark2)
    self.assertFalse(mark1 != mark2)
    self.assertTrue(set([mark1]) == set([mark1]))
    self.assertTrue(set([mark1]) == set([mark2]))


  def testOverrideTuple(self):
    mark1 = petri.Marking((0,0))
    mark2 = petri.Marking((0,0))
    self.assertTrue(mark1, mark2)
    self.assertTrue(mark1 <= mark2)
    self.assertTrue(mark1 >= mark2)
    self.assertFalse(mark1 < mark2)
    self.assertFalse(mark1 > mark2)

    mark1 = petri.Marking((0,0))
    mark2 = petri.Marking((1,0))
    self.assertNotEqual(mark1, mark2)
    self.assertTrue(mark1 <= mark2)
    self.assertFalse(mark1 >= mark2)
    self.assertTrue(mark1 < mark2)
    self.assertFalse(mark1 > mark2)

    mark1 = petri.Marking((1,1))
    mark2 = petri.Marking((2,0))
    self.assertNotEqual(mark1, mark2)
    self.assertFalse(mark1 <= mark2)
    self.assertFalse(mark1 >= mark2)
    self.assertFalse(mark1 < mark2)
    self.assertFalse(mark1 > mark2)

class SampleUnitTest(unittest.TestCase):

  def testSample(self):
    print
    good_ex = (ex for ex in sample.BuildExamples() if ex.name != 'mesh3x2' )
    map(self._testExample, good_ex)
    print '*'*40
    #print helper.Counter.nb_calls

  def _testExample(self, example):
    print "testing ", example.name, '\t',
    cover = petri.CoverSet(example.transition, example.initial)
    cover.Compute()
    if example.expected is not None:
      self.assertEqual(cover.ToSet(),  example.expected, example.name)

if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(SampleUnitTest)
    unittest.TextTestRunner().run(suite)
#unittest.main()

